-- A3Pv2.6/VectGen/VectGen.vhdl
-- created dim. déc.  8 08:36:48 CET 2019 by whygee@f-cpu.org
-- version dim. déc.  8 22:01:42 CET 2019 : gets the DUT's vector size.
--
-- Released under the GNU AGPLv3 license

Library ieee;
    use ieee.std_logic_1164.all;
    use ieee.numeric_std.all;
use STD.textio.all;
Library SLV;
    use SLV.SLV_utils.all;
Library work;
    use work.all;
Library proasic3;
    use proasic3.all;
    use proasic3.PA3_genlib.all;

entity VectGen is
  generic (
    VectGenWidthIn : integer := 200;
    VectGenWidthOut: integer := 100;
    main_delay : time := 1 sec;
    filename : string := ""; -- file containing the exclude vectors
    verbose : string := ""
  );
end VectGen;

architecture plip of VectGen is
  signal VectIn : std_logic_vector(VectGenWidthIn -1 downto 0) := (others=>'U');
  signal VectOut: std_logic_vector(VectGenWidthOut-1 downto 0) := (others=>'U');
  signal VectClk: std_logic := '0';

  shared variable VectInMax  : integer := VectGenWidthIn -1;
  shared variable VectOutMax : integer := VectGenWidthOut-2; -- -1 is used by VectInMax loopback

begin

  -- force the registration of the gates when placed at the very beginning
  rg: entity register_generics generic map(
         gate_select_number => 1, -- no listing
         bit_select_number => -1, -- no alteration
         flip_LUT => 2,           -- probe mode
         filename => filename,    -- eventual exclusions
         verbose => verbose)
  port map(clock => '0');   -- the updates are directly driven by main

  -- This code is executed after all the variables/signals are initialised
  main: process

    procedure GetVectorsSize is
    begin
      wait for main_delay;  -- let the circuit settle so we can read the output

      -- Find the size of the output vector
      FindOut: loop
        assert VectOutMax > 1
             report " Can't find the output vector size :-( "
             severity failure;
        exit when VectOut(VectOutMax) = '1';
        VectOutMax := VectOutMax - 1;
      end loop;

      -- Then find the size of the input vector
      FindIn: loop
        assert VectInMax > 1
          report " Can't find the intput vector size :-( "
          severity failure;
        VectIn(VectInMax) <= '1';
        wait for main_delay;
        exit when VectOut(VectGenWidthOut-1) = '1';
        VectInMax := VectInMax - 1;
      end loop;

      if verbose /= "" then
        report " Input vector : " & integer'image(VectInMax)
            & " bits, Output vector : " & integer'image(VectOutMax) & " bits";
      end if;

      -- avoid off-by-one
      VectInMax := VectInMax - 1;
      VectOutMax := VectOutMax - 1;

      -- allocate some memory for the descriptors
      output_records := new input_records_array(0 to VectOutMax);
       input_records := new output_record_array(0 to VectInMax);

    end GetVectorsSize;


    procedure GetNetList is
      variable i : integer;
      variable j : std_logic;

      procedure check_inputs is
      begin
--        if verbose /= "" then
--          report " \\\\\\\\\\\ BATCH TEST \\\\\\\\\\\ ";
--        end if;

        wait for main_delay;

        -- check the outputs too !
        for i in 0 to VectOutMax loop
          j := VectOut(i);
          if (j /= 'U') and (j /= 'X') and
            output_records(i).source = integer'low then
            output_records(i) := register_source(-i, 0, j, "output");
          end if;
        end loop;

        -- might no be so useless after all...
        probe_phase := 'X';
        VectIn(VectInMax downto 0) <= (others => probe_phase);
        wait for main_delay;

        -- erase the previous input vectors
        probe_phase := 'U';
        VectIn(VectInMax downto 0) <= (others => probe_phase);

--        if verbose /= "" then
--          report "////////  BATCH TEST END //////// ";
--        end if;
      end check_inputs;


    begin
      probe_phase := 'X';
      VectIn(VectInMax downto 0) <= (others => probe_phase);
      wait for main_delay;

      probe_phase := 'U';
      VectIn(VectInMax downto 0) <= (others => probe_phase);

      j := '0';
      probe_numbers := (others => integer'low);

      -- scan the inputs first :
      for i in 0 to VectInMax loop
--        if verbose /= "" then
--          report " Toggling input #" & integer'image(i) & "   _  " & std_logic'image(j);
--        end if;
        probe_numbers(j) := -i;  -- 0=>0, 1=>-1, 2=>-2...
        VectIn(i) <= j;
        if j = '-' then
          check_inputs;
          probe_numbers := (others => integer'low);
          j := '0';
        else
          j := std_logic'succ(j);
        end if;
      end loop;

      -- now scan the gates !
      for i in 1 to gate_instance_counter loop
--        if verbose /= "" then
--          report " Toggling output of gate #" & integer'image(i) & "   _  " & std_logic'image(j);
--        end if;
        probe_numbers(j) := i;  -- i >= 1

        if j = '-' then
          check_inputs;
          probe_numbers := (others => integer'low);
          j := '0';
        else
          j := std_logic'succ(j);
        end if;
      end loop;

      if j /= '0' then
        check_inputs;
      end if;

    end GetNetList;


    procedure FixDepthAndSinklist is
      variable i, j : integer;

      procedure list_sinks(driver: output_record_type) is
        variable g, n, o, p, g2, p2 : integer;
      begin
        g := driver.first_sink;
        n := driver.fanout;
        o := 1;
        p := driver.first_port;

        loop
          if g > 0 then
            report "     " & integer'image(o) & " : Gate #" & integer'image(g)
                 & "(" & integer'image(p) & ") - " & trim_name(GateList(g).gatename.all, ":dut@");
            g2 := GateList(g).in_record(p).next_input;
            p2 := GateList(g).in_record(p).next_port;
          else
            report "     " & integer'image(o) & " : Output #" & integer'image(-g);
            g2 := output_records(-g).next_input;
            p2 := output_records(-g).next_port;
          end if;
          n := n-1;
          exit when n < 1;
          o := o+1;
          g := g2;
          p := p2;
        end loop;
      end list_sinks;

    begin
      report " ************ FIXING NETLIST ************";

      -- scan the inputs first :
      for i in 0 to VectInMax loop
        j := input_records(i).fanout;
        if j > 0 then
          report "- Input #" & integer'image(i) & " : fanout=" & integer'image(j);
          list_sinks(input_records(i));
        else
          report "- Input #" & integer'image(i) & " : NO CONNECTION";
        end if;
      end loop;

      -- scan the gates
      for i in 1 to gate_instance_counter loop
        j := GateList(i).out_record.fanout;
        if j > 0 then
          report "- Gate #" & integer'image(i) & " : fanout=" & integer'image(j);
          list_sinks(GateList(i).out_record);
        else
          report "- Gate #" & integer'image(i) & " : NO CONNECTION";
	end if;
      end loop;

      report " ************ NETLIST FIXED ************";
    end FixDepthAndSinklist;


    procedure DisplayNetList is
      variable i, j : integer;

      procedure list_sinks(driver: output_record_type) is
        variable g, n, o, p, g2, p2 : integer;
      begin
        g := driver.first_sink;
        n := driver.fanout;
        o := 1;
        p := driver.first_port;

        loop
          if g > 0 then
            report "     " & integer'image(o) & " : Gate #" & integer'image(g)
                 & "(" & integer'image(p) & ") - " & trim_name(GateList(g).gatename.all, ":dut@");
            g2 := GateList(g).in_record(p).next_input;
            p2 := GateList(g).in_record(p).next_port;
          else
            report "     " & integer'image(o) & " : Output #" & integer'image(-g);
            g2 := output_records(-g).next_input;
            p2 := output_records(-g).next_port;
          end if;
          n := n-1;
          exit when n < 1;
          o := o+1;
          g := g2;
          p := p2;
        end loop;
      end list_sinks;

    begin
      report " ************ NETLIST ************";

      -- scan the inputs first :
      for i in 0 to VectInMax loop
        j := input_records(i).fanout;
        if j > 0 then
          report "- Input #" & integer'image(i) & " : fanout=" & integer'image(j);
          list_sinks(input_records(i));
        else
          report "- Input #" & integer'image(i) & " : NO CONNECTION";
        end if;
      end loop;

      -- scan the gates
      for i in 1 to gate_instance_counter loop
        j := GateList(i).out_record.fanout;
        if j > 0 then
          report "- Gate #" & integer'image(i) & " : fanout=" & integer'image(j);
          list_sinks(GateList(i).out_record);
        else
          report "- Gate #" & integer'image(i) & " : NO CONNECTION";
	end if;
      end loop;

      report " ************ END OF NETLIST ************";
    end DisplayNetList;


  begin
    GetVectorsSize;
    GetNetList;
    FixDepthAndSinklist;
--    DisplayNetList;
    wait;
  end process;

  -- Here we "plug" the wrapper to the Vector Generator ports.
  -- It must be the last section of the code so the drivers
  -- can initialise all the data behind the scene.
  wrap: entity VG_Wrapper
    generic map(
      VectGenWidthIn  => VectGenWidthIn,
      VectGenWidthOut => VectGenWidthOut)
    port map (
      VectIn  => VectIn,
      VectOut => VectOut,
      VectClk => VectClk);

end plip;
